# TPK4186 - Advanced Tools for Performance Engineering
# PERT.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# This work is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# Table of Contents
# -----------------
# 1. Required modules
# 2. Activity
# 3. Project

# 1. Required modules
# -------------------

import sys
import re

# 2. Activity
# -----------

class Activity:
  def __init__(self, name):
    self.name = name
    self.description = ""
    self.duration = 0
    self.predecessors = []
    self.successors = []
    self.earlyStartingDate = 0
    self.earlyCompletionDate = 0
    self.lateStartingDate = 0
    self.lateCompletionDate = 0

  def GetName(self):
    return self.name

  def GetDescription(self):
    return self.description

  def SetDescription(self, description):
    self.description = description

  def GetDuration(self):
    return self.duration

  def SetDuration(self, duration):
    self.duration = duration

  def GetPredecessors(self):
    return self.predecessors

  def GetSuccessors(self):
    return self.successors

  def AddPredecessor(self, activity):
    if activity in self.predecessors:
      return
    self.predecessors.append(activity)
    activity.successors.append(self)

  def GetEarlyStartingDate(self):
    return self.earlyStartingDate

  def SetEarlyStartingDate(self, date):
    self.earlyStartingDate = date

  def GetEarlyCompletionDate(self):
    return self.earlyCompletionDate

  def SetEarlyCompletionDate(self, date):
    self.earlyCompletionDate = date

  def GetLateStartingDate(self):
    return self.lateStartingDate

  def SetLateStartingDate(self, date):
    self.lateStartingDate = date

  def GetLateCompletionDate(self):
    return self.lateCompletionDate

  def SetLateCompletionDate(self, date):
    self.lateCompletionDate = date

  def IsCritical(self):
    return self.earlyStartingDate==self.lateStartingDate
 
# 3. Project
# -----------

class Project:
  def __init__(self):
    self.activities = []
    self.earlyCompletionDate = 0

  def GetActivities(self):
    return self.activities

  def LookForActivity(self, name):
    for activity in self.activities:
      if activity.GetName()==name:
        return activity
    return None

  def NewActivity(self, name):
    activity = self.LookForActivity(name)
    if activity==None:
      activity = Activity(name)
      self.activities.append(activity)
    return activity

  def RecordActivity(self, name, description, duration, predecessors):
    activity = self.NewActivity(name)
    activity.SetDescription(description)
    activity.SetDuration(duration)
    for predecessor in predecessors:
      activity.AddPredecessor(predecessor)
    return activity

  def GetEarlyCompletionDate(self):
    return self.earlyCompletionDate

  def ComputeEarlyDates(self):
    # Here we take benefit that if an activity A has been declared before an activity B
    # then B cannot precede A
    self.earlyCompletionDate = 0
    for activity in self.GetActivities():
      activity.SetEarlyStartingDate(0)
      for predecessor in activity.GetPredecessors():
        if predecessor.GetEarlyCompletionDate()>activity.GetEarlyStartingDate():
          activity.SetEarlyStartingDate(predecessor.GetEarlyCompletionDate())
      activity.SetEarlyCompletionDate(activity.GetEarlyStartingDate() + activity.GetDuration())
      if activity.GetEarlyCompletionDate()>self.earlyCompletionDate:
        self.earlyCompletionDate = activity.GetEarlyCompletionDate()

  def ComputeLateDates(self):
    activities = self.GetActivities()[:] # copy the list of activities
    activities.reverse()
    for activity in activities:
      activity.SetLateCompletionDate(self.earlyCompletionDate)
      for successor in activity.GetSuccessors():
        if successor.GetLateStartingDate()<activity.GetLateCompletionDate():
          activity.SetLateCompletionDate(successor.GetLateStartingDate())
      activity.SetLateStartingDate(activity.GetLateCompletionDate() - activity.GetDuration())

  def ExtractCriticalActivities(self):
    criticalActivities = []
    for activity in self.GetActivities():
      if activity.IsCritical():
        criticalActivities.append(activity)
    return criticalActivities
